// Byte-swapping, unaligned memory, and I/O port access functions. These
// functions are all unprivileged, except the ll_in_* and ll_out_*
// functions, whose privilege is architecture-dependent (on x64,
// they require IOPL 3).

#ifndef _LL_ARCH_IO_H
#define _LL_ARCH_IO_H

#include <lowlevel/types.h>

// Unconditional byte-swap

static inline uint16_t ll_swap16(uint16_t val)
{
	asm("xchgb %b0, %h0" : "+abcd" (val));
	return val;
}

static inline uint32_t ll_swap32(uint32_t val)
{
	asm("bswap %0" : "+r" (val));
	return val;
}

static inline uint64_t ll_swap64(uint64_t val)
{
	asm("bswap %0" : "+r" (val));
	return val;
}

// Conditional byte-swap to/from a specific endianness

static inline uint64_t ll_swap_le64(uint64_t val)
{
	return val;
}

static inline uint32_t ll_swap_le32(uint32_t val)
{
	return val;
}

static inline uint16_t ll_swap_le16(uint16_t val)
{
	return val;
}

static inline uint64_t ll_swap_be64(uint64_t val)
{
	return ll_swap64(val);
}

static inline uint32_t ll_swap_be32(uint32_t val)
{
	return ll_swap32(val);
}

static inline uint16_t ll_swap_be16(uint16_t val)
{
	return ll_swap16(val);
}

// Unaligned, host-endian

static inline uint64_t ll_load_unaligned_64(void *addr)
{
	return *(uint64_t *)addr;
}

static inline uint32_t ll_load_unaligned_32(void *addr)
{
	return *(uint32_t *)addr;
}

static inline uint16_t ll_load_unaligned_16(void *addr)
{
	return *(uint16_t *)addr;
}

static inline void ll_store_unaligned_64(void *addr, uint64_t val)
{
	*(uint64_t *)addr = val;
}

static inline void ll_store_unaligned_32(void *addr, uint32_t val)
{
	*(uint32_t *)addr = val;
}

static inline void ll_store_unaligned_16(void *addr, uint16_t val)
{
	*(uint16_t *)addr = val;
}

// Unaligned, big-endian

static inline uint16_t ll_load_unaligned_be16(void *addr)
{
	return ll_swap16(ll_load_unaligned_16(addr));
}

static inline uint32_t ll_load_unaligned_be32(void *addr)
{
	return ll_swap32(ll_load_unaligned_32(addr));
}

static inline uint64_t ll_load_unaligned_be64(void *addr)
{
	return ll_swap64(ll_load_unaligned_64(addr));
}

static inline void ll_store_unaligned_be16(void *addr, uint16_t val)
{
	ll_store_unaligned_16(addr, ll_swap16(val));
}

static inline void ll_store_unaligned_be32(void *addr, uint32_t val)
{
	ll_store_unaligned_32(addr, ll_swap32(val));
}

static inline void ll_store_unaligned_be64(void *addr, uint64_t val)
{
	ll_store_unaligned_64(addr, ll_swap64(val));
}

// Aligned, big-endian

static inline uint64_t ll_load_be64(uint64_t *addr)
{
	return ll_load_unaligned_be64(addr);
}

static inline uint32_t ll_load_be32(uint32_t *addr)
{
	return ll_load_unaligned_be32(addr);
}

static inline uint16_t ll_load_be16(uint16_t *addr)
{
	return ll_load_unaligned_be16(addr);
}

static inline void ll_store_be64(uint64_t *addr, uint64_t val)
{
	ll_store_unaligned_be64(addr, val);
}

static inline void ll_store_be32(uint32_t *addr, uint32_t val)
{
	ll_store_unaligned_be32(addr, val);
}

static inline void ll_store_be16(uint16_t *addr, uint16_t val)
{
	ll_store_unaligned_be16(addr, val);
}

// Unaligned, little-endian

static inline uint64_t ll_load_unaligned_le64(void *addr)
{
	return ll_load_unaligned_64(addr);
}

static inline uint32_t ll_load_unaligned_le32(void *addr)
{
	return ll_load_unaligned_32(addr);
}

static inline uint16_t ll_load_unaligned_le16(void *addr)
{
	return ll_load_unaligned_16(addr);
}

static inline void ll_store_unaligned_le32(void *addr, uint32_t val)
{
	ll_store_unaligned_32(addr, val);
}

static inline void ll_store_unaligned_le16(void *addr, uint16_t val)
{
	ll_store_unaligned_16(addr, val);
}

// Aligned, little-endian

static inline uint32_t ll_load_le32(uint32_t *addr)
{
	return ll_load_unaligned_le32(addr);
}

static inline uint16_t ll_load_le16(uint16_t *addr)
{
	return ll_load_unaligned_le16(addr);
}

static inline void ll_store_le32(uint32_t *addr, uint32_t val)
{
	ll_store_unaligned_le32(addr, val);
}

static inline void ll_store_le16(uint16_t *addr, uint16_t val)
{
	ll_store_unaligned_le16(addr, val);
}

// PCI/ISA/similar I/O-space; if memory-mapped, addr must include
// the base of the I/O window.

static inline uint8_t ll_in_8(uintptr_t addr)
{
	uint8_t val;
	asm volatile("inb %w1, %b0" : "=a" (val) : "d" (addr) : "memory");
	return val;
}

static inline uint16_t ll_in_le16(uintptr_t addr)
{
	uint16_t val;
	asm volatile("inw %w1, %w0" : "=a" (val) : "d" (addr) : "memory");
	return val;
}

static inline uint32_t ll_in_le32(uintptr_t addr)
{
	uint32_t val;
	asm volatile("inl %w1, %0" : "=a" (val) : "d" (addr) : "memory");
	return val;
}

static inline uint16_t ll_in_be16(uintptr_t addr)
{
	return ll_swap16(ll_in_le16(addr));
}

static inline uint32_t ll_in_be32(uintptr_t addr)
{
	return ll_swap32(ll_in_le32(addr));
}

static inline void ll_out_8(uintptr_t addr, uint8_t val)
{
	asm volatile("outb %b0, %w1" : : "a" (val), "d" (addr) : "memory");
}

static inline void ll_out_le16(uintptr_t addr, uint16_t val)
{
	asm volatile("outw %w0, %w1" : : "a" (val), "d" (addr) : "memory");
}

static inline void ll_out_le32(uintptr_t addr, uint32_t val)
{
	asm volatile("outl %0, %w1" : : "a" (val), "d" (addr) : "memory");
}

static inline void ll_out_be16(uintptr_t addr, uint16_t val)
{
	ll_out_le16(addr, ll_swap16(val));
}

static inline void ll_out_be32(uintptr_t addr, uint32_t val)
{
	ll_out_le32(addr, ll_swap32(val));
}

#endif
