#ifndef _KERN_ORB_H
#define _KERN_ORB_H

#include <kern/types.h>
#include <kern/radix.h>
#include <kern/lock.h>

#include <util/list.h>
#include <util/rbtree.h>
#include <util/radix.h>
#include <util/bmaptree.h>

#include <System/Objects.h>
#include <orb.h>

namespace Mem {
	class ProcAddrSpace;
};

namespace Threads {
	class Thread;
};

namespace ORB {
	typedef u32 ID;

	struct CallFrame {
		// Address Space and PC to return to
		Mem::ProcAddrSpace *ret_aspace;
		uintptr_t ret_stack;
		
		// Caller's PIB Pointer
		System::RunTime::ParamInfoBlock *caller_user_pib;
	};
	
	struct CallStackHeader {
		Threads::Thread *thread;
		Util::List node;
		
		// Number of CallFrames under this header.
		// For a full page, this is:
		//   (Arch::page_size - sizeof(CallStackHeader)) / sizeof(CallFrame)
		// There is also a much smaller group of call frames in the Thread
		// structure, so that an ORB stack page doesn't need to be allocated
		// in the common case of shallow method nesting.
	
		int num_frames;
		
		CallFrame frames[0];
	};

	struct ObjectHdr {
		ID id;

		union {
			struct {
				u32 Valid:1;
				u32 Pointer:1;
			};
			
			u32 flags;
		};
	};
	
	struct Object : public ObjectHdr {
		Mem::ProcAddrSpace *aspace;
	};

	struct ObjectPtr : public ObjectHdr {
		Object *object;
	};
	
	union ObjTableEntry {
		ObjectHdr hdr;
		Object obj;
		ObjectPtr ptr;
	};

	typedef Util::RadixTree<ObjTableEntry, ID, 6> IDTable;
	typedef Util::RBPtr<Object *, ObjectPtr *> IDRMap;
	typedef Util::BitmapTree<ID> IDAlloc;

	class IDSpace {
		Lock::SpinLock lock; // For add/del only; not needed for lookup
		IDTable table; // Forward mapping of local IDs to object ptrs
		IDRMap rmap; // Reverse mapping of remote object pointers to local IDs
		IDAlloc alloc; // Bitmap for allocating IDs
	
	public:
		Object *lookup(u32 id);
		ObjectHdr *get_local(Object *obj);
		Object *newobj(Mem::ProcAddrSpace *aspace);
	};
	
	void invoke_method(System::RunTime::ParamInfoBlock *pib, uintptr_t &stack);
	uintptr_t return_from_method(uintptr_t &exptr, size_t &exlen);
	void init();
}

#endif
