#ifndef _KERN_THREAD_H
#define _KERN_THREAD_H

#include <kern/types.h>
#include <kern/time.h>
#include <kern/orb.h>
#include <kern/sched.h>
#include <kern/lock.h>

#include <arch/thread.h>

#include <util/list.h>

namespace Mem {
	class ProcAddrSpace;
}

namespace Threads {
	class Thread {
		// This must be first.
		Arch::ArchThread arch;
		
		// If the thread is currently running, this contains the time of
		// the context switch when the thread most recently started
		// running.

		Time::Time last_time;
		
		// This contains the time in ns that the thread has remaining in
		// its current timeslice.
		
		s32 time_left;

		// The current timeslice length in ns.  For timeshared tasks, this
		// is simply the priority scaled linearly.
		
		u32 time_slice;
		
		// Real-time and timeshared priorities
		//
		// Real-time priorities go from 0 to 255, with 0 indicating a
		// timeshared task.
		//
		// Static timeshared priorities go from 1 to 16; they determine the
		// timeslice of a task (not including interactivity bonuses).  The
		// default is 8.  The dynamic timeshared priority is the priority
		// associated with time_left as of the last global timeslice
		// replenishment or wake-up, and ranges from 1 to 31.  If the
		// static priority is x, the dynamic priority is within the range
		// [x,2x-1].
		
		int rt_prio, ts_static_prio, ts_prio, last_replenish;
		
		enum Policy {
			TimeShared,
			FIFO,
			RoundRobin
		} policy;

		ThreadBlocker *blocked_on;
		Util::List runqueue_node;
		
		void ts_add();
		void ts_del();
		void ts_deplete();

		void add();
		void del();
		void replenish();
		void charge(Time::Time &now);
		void prio_adjust();
		void wake_nolock();
		
	public:
		Util::List threadlist_node;
		Mem::ProcAddrSpace *aspace, *active_aspace;
		
		enum {
			name_len = 32
		};
		
		char name[name_len];
	
		enum {
			internal_orbstack_frames = 8
		};
		
		ORB::CallStackHeader orbstack;
		ORB::CallFrame orbstackframes[internal_orbstack_frames];

		// The header and frame of the method invocation at the top of
		// the stack.  FIXME: add a lock aronud these if other threads
		// can read these to do a traceback/traceforward.

		ORB::CallStackHeader *orbstack_top_hdr;
		int orbstack_top; // Index into orbstack_top_hdr->frames[]
		
		~Thread();
		void exit();
		void block(ThreadBlocker *blocker);
		void wake();
		
		friend class Sched;
		friend class WaitQueue;

		friend void Arch::switch_thread(Thread *dest, Thread *src);
		
		// FIXME: temp hack; use a Process later
		
		void set_aspace(Mem::ProcAddrSpace *aspace);
	};
	
	enum {
		thread_size = sizeof(Thread) + 7 / 8
	};
}

#include <arch/current.h>
#define curthread (::Arch::get_current_thread())

namespace Lock {
	inline bool Lock::held_by_curthread()
	{
		return lockval == reinterpret_cast<ulong>(curthread);
	}
}

#endif
