// lib/kernel/time/timer.cc -- Generic timer multiplexing
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kernel/time.h>
#include <lowlevel/clock.h>

namespace Time {
	void TimerMux::arm(TimerEntry *entry, Time expiry)
	{
		AutoSpinLockRecIRQ autolock(lock);
		
		bool was_empty = heap.empty();
		Time oldexpiry;

		if (!was_empty) {
			oldexpiry = heap.get_top()->expiry;
		} else {
			// Shut up the compiler
			oldexpiry.seconds = 0;
			oldexpiry.nanos = 0;
		}

		entry->expiry = expiry;
		
		if (entry->is_armed())
			heap.requeue(entry);
		else
			heap.add(entry);

		Time newexpiry = heap.get_top()->expiry;
		if (was_empty || oldexpiry != newexpiry)
			parent.arm(newexpiry);
	}

	void TimerMux::disarm(TimerEntry *entry)
	{
		AutoSpinLockRecIRQ autolock(lock);
		
		if (entry->is_armed()) {
			bool was_top = entry == heap.get_top();

			heap.del(entry);

			if (was_top)
				parent.arm(heap.get_top()->expiry);
		}
	}
	
	void TimerMux::run()
	{
		DroppableAutoSpinLockRecIRQ autolock(lock);
		
		while (true) {
			TimerEntry *top = heap.get_top();
			Time now;
			clock.get_time(&now);

			if (now < top->expiry)
				break;

			heap.del(top);

			autolock.unlock();
			top->action();
			autolock.lock();
		}
		
		if (!heap.empty())
			parent.arm(heap.get_top()->expiry);
	}
	
	void TimerEntry::set_action(System::Events::Event *event)
	{
		// FIXME
	}

	void TimerEntry::action()
	{
		// FIXME
	}
}

#include _KERNEL_SERVER(time/timer/footer.cc)
